@@ -7,16 +7,19 @@
 # http://opensource.org/licenses/MIT>, at your option. This file may not be
 # copied, modified, or distributed except according to those terms.
 
+import base64
 import contextlib
 import datetime
+import json
 import mimetypes
 import os.path
+import sys
 
 import click
-import requests
-import urllib3
+import nethsm as nethsm_sdk
+from nethsm.backup import EncryptedBackup
 
-import pynitrokey.nethsm
+from pynitrokey.cli.exceptions import CliException
 from pynitrokey.helpers import prompt
 
 
@@ -33,15 +36,15 @@ KEY_CERTIFICATE_MIME_TYPES = [
 
 
 DATETIME_TYPE = click.DateTime(formats=["%Y-%m-%dT%H:%M:%S%z"])
-ROLE_TYPE = make_enum_type(pynitrokey.nethsm.Role)
-LOG_LEVEL_TYPE = make_enum_type(pynitrokey.nethsm.LogLevel)
-UNATTENDED_BOOT_STATUS_TYPE = make_enum_type(pynitrokey.nethsm.UnattendedBootStatus)
-TYPE_TYPE = make_enum_type(pynitrokey.nethsm.KeyType)
-TYPE_TLS_KEY_TYPE = make_enum_type(pynitrokey.nethsm.TlsKeyType)
-MECHANISM_TYPE = make_enum_type(pynitrokey.nethsm.KeyMechanism)
-ENCRYPT_MODE_TYPE = make_enum_type(pynitrokey.nethsm.EncryptMode)
-DECRYPT_MODE_TYPE = make_enum_type(pynitrokey.nethsm.DecryptMode)
-SIGN_MODE_TYPE = make_enum_type(pynitrokey.nethsm.SignMode)
+ROLE_TYPE = make_enum_type(nethsm_sdk.Role)
+LOG_LEVEL_TYPE = make_enum_type(nethsm_sdk.LogLevel)
+UNATTENDED_BOOT_STATUS_TYPE = make_enum_type(nethsm_sdk.UnattendedBootStatus)
+TYPE_TYPE = make_enum_type(nethsm_sdk.KeyType)
+TYPE_TLS_KEY_TYPE = make_enum_type(nethsm_sdk.TlsKeyType)
+MECHANISM_TYPE = make_enum_type(nethsm_sdk.KeyMechanism)
+ENCRYPT_MODE_TYPE = make_enum_type(nethsm_sdk.EncryptMode)
+DECRYPT_MODE_TYPE = make_enum_type(nethsm_sdk.DecryptMode)
+SIGN_MODE_TYPE = make_enum_type(nethsm_sdk.SignMode)
 
 
 def print_row(values, widths):
@@ -91,9 +94,6 @@ def nethsm(ctx, host, version, username, password, verify_tls):
     ctx.obj["NETHSM_PASSWORD"] = password
     ctx.obj["NETHSM_VERIFY_TLS"] = verify_tls
 
-    if not verify_tls:
-        urllib3.disable_warnings()
-
 
 @contextlib.contextmanager
 def connect(ctx, require_auth=True):
@@ -113,26 +113,20 @@ def connect(ctx, require_auth=True):
                 f"[auth] Password for user {username} on NetHSM {host}", hide_input=True
             )
 
-    with pynitrokey.nethsm.connect(
-        host, version, username, password, verify_tls
-    ) as nethsm:
-        import urllib3.exceptions
-
+    with nethsm_sdk.connect(host, version, username, password, verify_tls) as nethsm:
         try:
             yield nethsm
-        except pynitrokey.nethsm.NetHSMError as e:
+        except nethsm_sdk.NetHSMError as e:
             raise click.ClickException(e)
-        except urllib3.exceptions.MaxRetryError as e:
-            if isinstance(e.reason, urllib3.exceptions.SSLError):
+        except nethsm_sdk.NetHSMRequestError as e:
+            if e.type == nethsm_sdk.RequestErrorType.SSL_ERROR:
                 raise click.ClickException(
                     f"Could not connect to the NetHSM: {e.reason}\nIf you use a self-signed certificate, please set the --no-verify-tls option."
                 )
             else:
-                raise e
-        except requests.exceptions.SSLError as e:
-            raise click.ClickException(
-                f"Could not connect to the NetHSM: {e}\nIf you use a self-signed certificate, please set the --no-verify-tls option."
-            )
+                raise click.ClickException(
+                    f"Cound not connect to the NetHSM: {e.reason}\nIs the NetHSM running and reachable?"
+                )
 
 
 @nethsm.command()
@@ -517,14 +511,14 @@ def delete_key(ctx, key_id):
 def prompt_mechanisms(type):
     # We assume that key type X corresponds to the mechanisms starting with X.
     # This is no longer true for curves, so we have to adapt the type
-    if type == pynitrokey.nethsm.KeyType.CURVE25519.value:
+    if type == nethsm_sdk.KeyType.CURVE25519.value:
         type = "EdDSA"
     elif type.startswith("EC_"):
         type = "ECDSA"
 
     available_mechanisms = []
     print("Supported mechanisms for this key type:")
-    for mechanism in pynitrokey.nethsm.KeyMechanism:
+    for mechanism in nethsm_sdk.KeyMechanism:
         if mechanism.value.startswith(type):
             available_mechanisms.append(mechanism.value)
             print(f"  {mechanism.value}")
@@ -1138,6 +1132,10 @@ def backup(ctx, filename):
         with open(filename, "xb") as f:
             f.write(data)
             print(f"Backup for {nethsm.host} written to {filename}")
+        try:
+            EncryptedBackup.parse(data)
+        except ValueError as e:
+            raise CliException(f"Failed to validate backup: {e}", support_hint=False)
 
 
 @nethsm.command()
@@ -1154,20 +1152,115 @@ def backup(ctx, filename):
     type=DATETIME_TYPE,
     help="The system time to set (default: the time of this system)",
 )
+@click.option(
+    "-f",
+    "--force",
+    is_flag=True,
+    help="Restore the backup even if validation fails",
+)
 @click.argument("filename")
 @click.pass_context
-def restore(ctx, backup_passphrase, system_time, filename):
+def restore(ctx, backup_passphrase, system_time, force, filename):
     """Restore a backup of a NetHSM instance from a file.
 
     If the system time is not set, the current system time is used."""
     if not system_time:
         system_time = datetime.datetime.now(datetime.timezone.utc)
+
+    with open(filename, "rb") as f:
+        data = f.read()
+    try:
+        EncryptedBackup.parse(data).decrypt(backup_passphrase)
+    except ValueError as e:
+        if force:
+            print(f"Failed to validate backup: {e}")
+            print("Backup is restored anyway as --force is set")
+        else:
+            raise CliException(
+                f"Failed to validate backup (use --force to restore anyway): {e}",
+                support_hint=False,
+            )
+
     with connect(ctx, require_auth=False) as nethsm:
         with open(filename, "rb") as f:
-            nethsm.restore(f, backup_passphrase, system_time)
+            nethsm.restore(data, backup_passphrase, system_time)
         print(f"Backup restored on NetHSM {nethsm.host}")
 
 
+@nethsm.command()
+@click.option(
+    "-p",
+    "--backup-passphrase",
+    help="The backup passphrase for decryption (default: only the unencrypted metadata is validated)",
+)
+@click.argument("filename")
+def validate_backup(backup_passphrase, filename):
+    """Validate a NetHSM backup file.
+
+    Per default, only the metadata of the encrypted backup is validated.  If
+    the backup passphrase is set, the backup is decrypted and the content is
+    also validated."""
+
+    with open(filename, "rb") as f:
+        data = f.read()
+    try:
+        encrypted = EncryptedBackup.parse(data)
+    except ValueError as e:
+        raise CliException(
+            f"Failed to validate backup metadata: {e}", support_hint=False
+        )
+
+    if backup_passphrase:
+        try:
+            encrypted.decrypt(backup_passphrase)
+        except ValueError as e:
+            raise CliException(
+                f"Failed to validate backup content: {e}", support_hint=False
+            )
+        print("Backup metadata and content are valid.")
+    else:
+        print("Backup metadata is valid.")
+
+
+@nethsm.command()
+@click.option(
+    "-p",
+    "--backup-passphrase",
+    hide_input=True,
+    prompt=True,
+    help="The backup passphrase",
+)
+@click.argument("filename")
+def export_backup(backup_passphrase, filename):
+    """Export the content of a NetHSM backup file.
+
+    The key-value data stored in the backup file is printed to the standard
+    output as a JSON object using the base64 encoding for binary data.
+    Additionally, the .locked-domain-key and .version keys are set with the
+    domain key and version info extracted from the backup file."""
+
+    with open(filename, "rb") as f:
+        data = f.read()
+    try:
+        encrypted = EncryptedBackup.parse(data)
+    except ValueError as e:
+        raise CliException(f"Failed to parse backup metadata: {e}", support_hint=False)
+
+    try:
+        decrypted = encrypted.decrypt(backup_passphrase)
+    except ValueError as e:
+        raise CliException(f"Failed to decrypt backup content: {e}", support_hint=False)
+
+    data = {}
+    data[".locked-domain-key"] = base64.b64encode(decrypted.domain_key).decode()
+    data[".version"] = decrypted.version
+    for key, value in decrypted.data.items():
+        data[key] = base64.b64encode(value).decode()
+
+    json.dump(data, sys.stdout, indent=4)
+    print()
+
+
 @nethsm.command()
 @click.argument("filename")
 @click.pass_context
@@ -1341,13 +1434,21 @@ def encrypt(ctx, key_id, data, mode, iv):
     prompt=True,
     help="The decrypt mode",
 )
+@click.option(
+    "-iv",
+    "--initialization-vector",
+    "iv",
+    type=str,
+    default="",
+    help="The initialization vector",
+)
 @click.pass_context
-def decrypt(ctx, key_id, data, mode):
+def decrypt(ctx, key_id, data, mode, iv):
     """Decrypt data with a secret key on the NetHSM and print the decrypted message.
 
     This command requires authentication as a user with the Operator role."""
     with connect(ctx) as nethsm:
-        print(nethsm.decrypt(key_id, data, mode))
+        print(nethsm.decrypt(key_id, data, mode, iv))
 
 
 @nethsm.command()
