@@ -12,14 +12,17 @@ from enum import Enum, IntEnum
 from hashlib import pbkdf2_hmac
 from secrets import token_bytes
 from struct import pack
-from typing import List, Optional
+from typing import Any, Callable, List, Optional, Sequence, Tuple, Union
 
 import semver
 import tlv8
 
-from pynitrokey.nk3 import Nitrokey3Device
+from pynitrokey.nk3.device import Nitrokey3Device
 from pynitrokey.start.gnuk_token import iso7816_compose
 
+LogFn = Callable[[str], Any]
+WriteCorpusFn = Callable[[typing.Union["Instruction", "CCIDInstruction"], bytes], Any]
+
 
 @dataclasses.dataclass
 class ListItemProperties:
@@ -75,7 +78,7 @@ class PasswordSafeEntry:
     properties: Optional[bytes] = None
     name: Optional[bytes] = None
 
-    def tlv_encode(self) -> List[tlv8.Entry]:
+    def tlv_encode(self) -> list[tlv8.Entry]:
         entries = [
             tlv8.Entry(Tag.PwsLogin.value, self.login)
             if self.login is not None
@@ -94,7 +97,7 @@ class PasswordSafeEntry:
 
 @dataclasses.dataclass
 class RawBytes:
-    data: List
+    data: list[int]
 
 
 @dataclasses.dataclass
@@ -216,6 +219,7 @@ class Instruction(Enum):
     ChangePIN = 0xB3
     SetPIN = 0xB4
     GetCredential = 0xB5
+    UpdateCredential = 0xB7
 
 
 class Tag(Enum):
@@ -291,28 +295,30 @@ class SecretsApp:
     """
 
     log: logging.Logger
-    logfn: typing.Callable
+    logfn: LogFn
     dev: Nitrokey3Device
-    write_corpus_fn: Optional[typing.Callable]
+    write_corpus_fn: Optional[WriteCorpusFn]
     _cache_status: Optional[SelectResponse]
-    _metadata: dict
+    _metadata: dict[Any, Any]
 
-    def __init__(self, dev: Nitrokey3Device, logfn: Optional[typing.Callable] = None):
+    def __init__(self, dev: Nitrokey3Device, logfn: Optional[LogFn] = None):
         self._cache_status = None
         self.write_corpus_fn = None
         self.log = logging.getLogger("otpapp")
         if logfn is not None:
-            self.logfn = logfn  # type: ignore [assignment]
+            self.logfn = logfn
         else:
-            self.logfn = self.log.info  # type: ignore [assignment]
+            self.logfn = self.log.info
         self.dev = dev
         self._metadata = {}
 
-    def _custom_encode(self, structure: Optional[List] = None) -> bytes:
+    def _custom_encode(
+        self, structure: Optional[Sequence[Union[tlv8.Entry, RawBytes, None]]] = None
+    ) -> bytes:
         if not structure:
             return b""
 
-        def transform(d: typing.Union[tlv8.Entry, RawBytes, None]) -> bytes:
+        def transform(d: Union[tlv8.Entry, RawBytes, None]) -> bytes:
             if not d:
                 return b""
             if isinstance(d, RawBytes):
@@ -331,7 +337,7 @@ class SecretsApp:
     def _send_receive(
         self,
         ins: typing.Union[Instruction, CCIDInstruction],
-        structure: Optional[List] = None,
+        structure: Optional[Sequence[Union[tlv8.Entry, RawBytes, None]]] = None,
     ) -> bytes:
         encoded_structure = self._custom_encode(structure)
         ins_b, p1, p2 = self._encode_command(ins)
@@ -366,7 +372,7 @@ class SecretsApp:
                 )
             log_multipacket = True
             ins_b, p1, p2 = self._encode_command(Instruction.SendRemaining)
-            bytes_data = iso7816_compose(ins_b, p1, p2, data=[])
+            bytes_data = iso7816_compose(ins_b, p1, p2)
             try:
                 result = self.dev.otp(data=bytes_data)
             except Exception as e:
@@ -422,9 +428,7 @@ class SecretsApp:
         self.logfn("Executing reset")
         self._send_receive(Instruction.Reset)
 
-    def list(
-        self, extended: bool = False
-    ) -> typing.List[typing.Union[typing.Tuple[bytes, bytes], bytes]]:
+    def list(self, extended: bool = False) -> list[Union[Tuple[bytes, bytes], bytes]]:
         """
         Return a list of the registered credentials
         :return: List of bytestrings, or tuple of bytestrings, if "extended" switch is provided
@@ -432,7 +436,7 @@ class SecretsApp:
         """
         raw_res = self._send_receive(Instruction.List)
         resd: tlv8.EntryList = tlv8.decode(raw_res)
-        res: List[typing.Union[typing.Tuple[bytes, bytes], bytes]] = []
+        res: list[Union[Tuple[bytes, bytes], bytes]] = []
         for e in resd:
             # e: tlv8.Entry
             if extended:
@@ -488,6 +492,52 @@ class SecretsApp:
         p.properties = p.properties.hex().encode() if p.properties else None
         return p
 
+    def rename_credential(self, cred_id: bytes, new_name: bytes) -> None:
+        """
+        Rename credential.
+        An alias for the update_credential() call.
+        @param cred_id: The credential ID to modify
+        @param new_name: New ID for the credential
+        """
+        return self.update_credential(cred_id, new_name)
+
+    def update_credential(
+        self,
+        cred_id: bytes,
+        new_name: Optional[bytes] = None,
+        login: Optional[bytes] = None,
+        password: Optional[bytes] = None,
+        metadata: Optional[bytes] = None,
+        touch_button: Optional[bool] = None,
+    ) -> None:
+        """
+        Update credential fields - name, attributes, and PWS fields.
+        Unpopulated fields will not be encoded and used during the update process
+        (won't change the current value).
+        @param cred_id: The credential ID to modify
+        @param new_name: New ID for the credential
+        @param login: New login field content
+        @param password: New password field content
+        @param metadata: New metadata field content
+        @param touch_button: Set if the touch button use should be required
+        """
+        structure = [
+            tlv8.Entry(Tag.CredentialId.value, cred_id),
+            tlv8.Entry(Tag.CredentialId.value, new_name) if new_name else None,
+            self.encode_properties_to_send(touch_button, False, tlv=True)
+            if touch_button is not None
+            else None,
+            tlv8.Entry(Tag.PwsLogin.value, login) if login is not None else None,
+            tlv8.Entry(Tag.PwsPassword.value, password)
+            if password is not None
+            else None,
+            tlv8.Entry(Tag.PwsMetadata.value, metadata)
+            if metadata is not None
+            else None,
+        ]
+        structure = list(filter(lambda x: x is not None, structure))
+        self._send_receive(Instruction.UpdateCredential, structure=structure)
+
     def delete(self, cred_id: bytes) -> None:
         """
         Delete credential with the given id. Does not fail, if the given credential does not exist.
@@ -558,13 +608,7 @@ class SecretsApp:
             tlv8.Entry(
                 Tag.Key.value, bytes([kind.value | algo.value, digits]) + secret
             ),
-            RawBytes(
-                [
-                    Tag.Properties.value,
-                    (0x02 if touch_button_required else 0x00)
-                    | (0x04 if pin_based_encryption else 0x00),
-                ]
-            ),
+            self.encode_properties_to_send(touch_button_required, pin_based_encryption),
             tlv8.Entry(
                 Tag.InitialCounter.value, initial_counter_value.to_bytes(4, "big")
             )
@@ -577,6 +621,25 @@ class SecretsApp:
         structure = list(filter(lambda x: x is not None, structure))
         self._send_receive(Instruction.Put, structure)
 
+    @classmethod
+    def encode_properties_to_send(
+        cls, touch_button_required: bool, pin_based_encryption: bool, tlv: bool = False
+    ) -> RawBytes:
+        """
+        Encode properties structure into a single byte
+        @param touch_button_required: whether the touch button use is required
+        @param pin_based_encryption: whether the PIN-encryption is requested (only during registration)
+        @param tlv: set True, if this should be encoded as TLV, as opposed to the default "TV", w/o L
+        """
+        structure = [
+            Tag.Properties.value,
+            1 if tlv else None,
+            (0x02 if touch_button_required else 0x00)
+            | (0x04 if pin_based_encryption else 0x00),
+        ]
+        structure = list(filter(lambda x: x is not None, structure))
+        return RawBytes(structure)  # type: ignore[arg-type]
+
     def calculate(self, cred_id: bytes, challenge: Optional[int] = None) -> bytes:
         """
         Calculate the OTP code for the credential named `cred_id`, and with challenge `challenge`.
@@ -699,7 +762,7 @@ class SecretsApp:
         ]
         raw_res = self._send_receive(Instruction.Validate, structure=structure)
         resd: tlv8.EntryList = tlv8.decode(raw_res)
-        return resd.data
+        return resd.data  # type: ignore[no-any-return]
 
     def select(self) -> SelectResponse:
         """
