@@ -11,6 +11,7 @@ import enum
 import logging
 import platform
 import re
+import time
 from abc import ABC, abstractmethod
 from contextlib import contextmanager
 from io import BytesIO
@@ -20,7 +21,7 @@ from spsdk.mboot.exceptions import McuBootConnectionError
 
 import pynitrokey
 from pynitrokey.helpers import Retries
-from pynitrokey.nk3 import Nitrokey3Base
+from pynitrokey.nk3.base import Nitrokey3Base
 from pynitrokey.nk3.bootloader import (
     FirmwareContainer,
     Nitrokey3Bootloader,
@@ -324,6 +325,9 @@ class Updater:
                     "The reboot was not confirmed with the touch button"
                 )
 
+            # needed for udev to properly handle new device
+            time.sleep(1)
+
             if platform.system() == "Darwin":
                 # Currently there is an issue with device enumeration after reboot on macOS, see
                 # <https://github.com/Nitrokey/pynitrokey/issues/145>.  To avoid this issue, we
