@@ -9,6 +9,7 @@ import hashlib
 import hmac
 import logging
 import time
+from contextlib import suppress
 from datetime import timedelta
 from os import environ, wait
 from sys import stderr
@@ -23,6 +24,7 @@ from pynitrokey.conftest import (
     CHALLENGE,
     CHALLENGE_RESPONSE_COMMANDS,
     CREDID,
+    CREDID2,
     DELAY_AFTER_FAILED_REQUEST_SECONDS,
     DIGITS,
     FEATURE_BRUTEFORCE_PROTECTION_ENABLED,
@@ -286,11 +288,13 @@ def test_reverse_hotp_failure(secretsAppRaw):
         assert not secretsApp.verify_code(CREDID, 10**5)
     helper_wait_after_failed_hotp_verification_request()
 
+    secretsApp.delete(CREDID)
     secretsApp.register(CREDID, secretb, digits=7, kind=Kind.Hotp, algo=Algorithm.Sha1)
     with pytest.raises(SecretsAppException, match="ConditionsOfUseNotSatisfied"):
         assert not secretsApp.verify_code(CREDID, 10**6)
     helper_wait_after_failed_hotp_verification_request()
 
+    secretsApp.delete(CREDID)
     secretsApp.register(CREDID, secretb, digits=8, kind=Kind.Hotp, algo=Algorithm.Sha1)
     with pytest.raises(SecretsAppException, match="ConditionsOfUseNotSatisfied"):
         assert not secretsApp.verify_code(CREDID, 10**7)
@@ -640,8 +644,10 @@ def test_too_long_message2(secretsAppRaw):
     for i in range(1, 255, 5):
         secretsApp.logfn(f"Testing secret length {i} bytes")
         try:
-            secretsApp.register("C", too_long_name[:i], DIGITS, kind=Kind.Hotp)
-            codes.append(secretsApp.calculate("C"))
+            with suppress(SecretsAppException):
+                secretsAppRaw.delete(CREDID)
+            secretsApp.register(CREDID, too_long_name[:i], DIGITS, kind=Kind.Hotp)
+            codes.append(secretsApp.calculate(CREDID))
         except Exception:
             break
     assert i >= 40, "Maximum secret length should be at least 320 bits"
@@ -1093,7 +1099,7 @@ def test_check_commands_access(secretsAppRaw):
             ):
                 secretsAppRaw._send_receive(ins, structure)
         else:
-            # These commands do not accept arguments, so there is no parsing error
+            # These commands do not expect arguments, so there is no parsing error
             secretsAppRaw._send_receive(ins, structure)
 
 
@@ -1494,12 +1500,17 @@ def test_hmac_low_level(secretsAppRaw):
         "1c e3 0f d7 8d 20 dc fa 40 b5 0c 18 77 9a fb 0f 02 28 8d b7".replace(" ", "")
     )
     for slot_name in [b"HmacSlot2", b"HmacSlot1"]:
+        with suppress(SecretsAppException):
+            secretsAppRaw.delete(slot_name)
         secretsAppRaw.register(
             slot_name,
             secret=secret,
             kind=Kind.Hmac,
         )
 
+    with suppress(SecretsAppException):
+        secretsAppRaw.delete(slot_name)
+
     # Do not allow to register secret with different lengths than expected 20 bytes
     for secret_len in [18, 21, 200]:
         with pytest.raises(SecretsAppException, match="IncorrectDataParameter"):
@@ -1534,13 +1545,27 @@ def test_hmac_low_level(secretsAppRaw):
      */
     """
 
-    # The length of "1" is used by KeepassXC for test purposes. "63" is the maximum.
+    # Prepare HmacSlot2 again
+    slot_name = b"HmacSlot2"
+    with suppress(SecretsAppException):
+        secretsAppRaw.delete(slot_name)
+    secretsAppRaw.register(
+        slot_name,
+        secret=secret,
+        kind=Kind.Hmac,
+    )
+
+    # The length of "1" is used by KeepassXC for the purposes. "63" is the maximum.
     # "64" should not work, as the last byte is always treated as the padding byte value.
     for challenge_len in [1, 32, 63]:
         challenge = b"c" * challenge_len
         challenge_padded = helper_get_padded(challenge)
         status, response_device = helper_send_receive_ins(
-            secretsAppRaw, YK_API_REQ, p1=slot, le=20, data_raw=challenge_padded
+            secretsAppRaw,
+            YK_API_REQ,
+            p1=YK_P1_CMD_HMAC_2,
+            le=20,
+            data_raw=challenge_padded,
         )
         response_lib = secretsAppRaw.get_response_for_secret(challenge, secret)
         assert response_lib == response_device
@@ -1651,3 +1676,180 @@ def test_light_load(secretsAppRaw):
         password=b"password",
         metadata=b"metadata",
     )
+
+
+@pytest.mark.parametrize(
+    "cred1_encryption", [True, False], ids=lambda x: "cred1" + ("_enc" if x else "")
+)
+@pytest.mark.parametrize(
+    "cred2_encryption", [True, False], ids=lambda x: "cred2" + ("_enc" if x else "")
+)
+def test_register_overwrite_attempt(secretsAppRaw, cred1_encryption, cred2_encryption):
+    """
+    Credential should not be overwritten with another one, regardless whether it is PIN-encrypted or not
+    """
+    app = secretsAppRaw
+    app.reset()
+    app.set_pin_raw(PIN)
+
+    if cred1_encryption:
+        app.verify_pin_raw(PIN)
+    app.register(
+        CREDID, SECRET, DIGITS, pin_based_encryption=cred1_encryption, kind=Kind.Hotp
+    )
+    if cred2_encryption:
+        app.verify_pin_raw(PIN)
+    with pytest.raises(SecretsAppException, match="OperationBlocked"):
+        app.register(
+            CREDID,
+            SECRET,
+            DIGITS,
+            pin_based_encryption=cred2_encryption,
+            kind=Kind.Hotp,
+        )
+
+
+def test_rename_credential(secretsAppResetLogin):
+    """
+    Credential should change its name. Test both PIN- and HW-encrypted credentials.
+    """
+    app = secretsAppResetLogin
+    app.register(CREDID, SECRET, DIGITS, kind=Kind.Hotp)
+    app.verify_pin_raw(PIN)
+    l = app.list()
+    assert len(l) == 1
+    assert l[0].decode() == CREDID
+    app.verify_pin_raw(PIN)
+    app.rename_credential(CREDID, CREDID2)
+    app.verify_pin_raw(PIN)
+    # There should be only 1 credential left, with a new name
+    l = app.list()
+    assert len(l) == 1
+    assert l[0].decode() == CREDID2
+
+    # Old name should not be accessible
+    app.verify_pin_raw(PIN)
+    with pytest.raises(SecretsAppException, match="NotFound"):
+        app.get_credential(CREDID)
+
+
+@pytest.mark.parametrize(
+    "cred1_encryption", [True, False], ids=lambda x: "cred1" + ("_enc" if x else "")
+)
+@pytest.mark.parametrize(
+    "cred2_encryption", [True, False], ids=lambda x: "cred2" + ("_enc" if x else "")
+)
+def test_rename_credential_to_existing(
+    secretsAppRaw, cred2_encryption, cred1_encryption
+):
+    """
+    Credential should not change its name to an existing one, regardless if the other is PIN-encrypted or not
+    """
+    app = secretsAppRaw
+    app.reset()
+    app.set_pin_raw(PIN)
+
+    if cred2_encryption:
+        app.verify_pin_raw(PIN)
+    app.register(
+        CREDID2, SECRET, DIGITS, kind=Kind.Hotp, pin_based_encryption=cred2_encryption
+    )
+    app.verify_pin_raw(PIN)
+    assert app.list()[0].decode() == CREDID2
+
+    if cred1_encryption:
+        app.verify_pin_raw(PIN)
+    app.register(
+        CREDID, SECRET, DIGITS, kind=Kind.Hotp, pin_based_encryption=cred1_encryption
+    )
+    app.verify_pin_raw(PIN)
+    # Once set up, there should be 2 credentials
+    assert set([CREDID.encode(), CREDID2.encode()]) == set(app.list())
+    app.verify_pin_raw(PIN)
+    # The rename operation should fail
+    with pytest.raises(SecretsAppException, match="OperationBlocked"):
+        app.rename_credential(CREDID, CREDID2)
+    app.verify_pin_raw(PIN)
+    # There should be still 2 credentials left
+    assert set([CREDID.encode(), CREDID2.encode()]) == set(app.list())
+
+
+def test_update_credential(secretsAppResetLogin):
+    """
+    Credential should change its properties. Test both PIN- and HW-encrypted credentials.
+    """
+    app = secretsAppResetLogin
+    app.register(CREDID, SECRET, DIGITS, kind=Kind.Hotp)
+    app.verify_pin_raw(PIN)
+    l = app.list_with_properties()
+    assert len(l) == 1
+    assert l[0].label.decode() == CREDID
+    assert not l[0].properties.touch_required
+    app.verify_pin_raw(PIN)
+    app.update_credential(CREDID, touch_button=True)
+    # There should be only one credential left, with the same name
+    app.verify_pin_raw(PIN)
+    l = app.list_with_properties()
+    assert len(l) == 1
+    assert l[0].label.decode() == CREDID
+    # This credential should be listed now as requiring a touch button for use
+    assert l[0].properties.touch_required
+
+    # Now add some PWS fields, and rename it too
+    app.verify_pin_raw(PIN)
+    c = app.get_credential(CREDID)
+    assert not c.login
+    assert not c.password
+    assert not c.metadata
+    app.verify_pin_raw(PIN)
+    app.update_credential(
+        CREDID,
+        new_name=CREDID2,
+        login=b"login",
+        password=b"password",
+        metadata=b"metadata",
+    )
+
+    # Check if PWS fields are there, and the "touch button required" flag is still present
+    app.verify_pin_raw(PIN)
+    c = app.get_credential(CREDID2)
+    assert c.login == b"login"
+    assert c.password == b"password"
+    assert c.metadata == b"metadata"
+    app.verify_pin_raw(PIN)
+    l = app.list_with_properties()
+    assert len(l) == 1
+    assert l[0].properties.touch_required
+
+    # Old name should not be accessible
+    app.verify_pin_raw(PIN)
+    with pytest.raises(SecretsAppException, match="NotFound"):
+        app.get_credential(CREDID)
+
+    # Try to remove the PWS data with empty strings, and rename again
+    app.verify_pin_raw(PIN)
+    app.update_credential(
+        CREDID2, new_name=CREDID, login=b"", password=b"", metadata=b""
+    )
+    app.verify_pin_raw(PIN)
+    c = app.get_credential(CREDID)
+    assert c.login is None
+    assert c.password is None
+    assert c.metadata is None
+
+    # Disallow to register a PWS credential with any 0-length strings field
+    app.verify_pin_raw(PIN)
+    with pytest.raises(SecretsAppException, match="IncorrectDataParameter"):
+        app.register(
+            CREDID2,
+            SECRET,
+            DIGITS,
+            kind=Kind.Hotp,
+            login=b"",
+            password=b"",
+            metadata=b"",
+        )
+    for i in ["login", "password", "metadata"]:
+        fields = {i: b""}
+        with pytest.raises(SecretsAppException, match="IncorrectDataParameter"):
+            app.register(CREDID2, SECRET, DIGITS, kind=Kind.Hotp, **fields)
